
# patPlectrum.py

"""
This module is an integeral part of the program
MMA - Musical Midi Accompaniment.

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

Bob van der Poel <bob@mellowood.ca>
Louis James Barman

"""


import MMA.notelen
import MMA.harmony

import gbl
from   MMA.common import *
from   MMA.pat import PC


class Plectrum(PC):
    """ Pattern class for a Raw MIDI track. """


    vtype = 'PLECTRUM'
    # We have vibrating strings (a string in python refers to text not a guitar string)
    _vibrating = []
    _tuning = []
    capoFretNo = 0  # The number that the capo is on (0 for open strings)

    def __init__(self, nm):
        PC.__init__(self, nm)
        self.setTuning("e-- a-- d- g- b- e")


    def addString(self, pitchName):
        midiPitch = noteNameToMidiPitch(pitchName)

        self._tuning.append(midiPitch)
        vibration = struct()
        vibration.note = -1       # -1 means the string not vibrating
        self._vibrating.append(vibration)


    def setTuning (self, stringPitchNames):
        """ for standard guitar tuning use setTuning("e-- a-- d- g- b- e") """

        #clear the old tuning
        del self._tuning[:]
        del self._vibrating[:]
        for pitchName in stringPitchNames.split():
            self.addString(pitchName)


    def getPgroup(self, ev):
        """ Get group for rawmid pattern.

            Fields - start, length, note, volume

        """

        if not (len(ev) == 3 or len(ev) == len(self._tuning) + 2 ):
            error("There must be n groups of 3 or %d in a pattern definition, "
                  "not <%s>" % (len(self._tuning) + 2, ' '.join(ev)) )

        a = struct()

        a.offset   = self.setBarOffset(ev[0])
        a.strum = stoi(ev[1], "Strum in Plectrum definition not int")

        a.pluckVol = []
        if len(ev) == 3:
            vol = stoi(ev[2], "Note volume in Plectrum definition not int")
            for stringNo in range(len(self._tuning)):
                a.pluckVol.append(vol)
        else:
            for stringNo in range(len(self._tuning)):
                val = ev[2 + stringNo]
                if val=='-':
                    val = '-1'   # -1 means ignore this string
                a.pluckVol.append(stoi(val, "Note volume in Plectrum definition not int"))

        return a




    def restart(self):
        self.ssvoice = -1




    def endVibration(self, stringNo, offset):
        """ kill the vibration on the string by sending out a note off """

        # first test if this srting has been played before
        if  self._vibrating[stringNo].note == -1:
            return

        vibration = self._vibrating[stringNo]
        if gbl.debug:
            print "    endVibration channel ", self.channel, "offset", offset, "note off ", vibration.note
        gbl.mtrks[self.channel].addNoteOnToTrack( offset, vibration.note, 0) # when v=0 this is a note off
        self._vibrating[stringNo].note = -1



    # endAllVibrations (ie output all outstanding note off)
    def grooveFinish(self, offset):
        print "grooveFinish offset=", offset, "self.channel=", self.channel, "gbl.mtrks[self.channel].channel=", gbl.mtrks[self.channel].channel
        if gbl.mtrks[self.channel].channel == -1:
            return
        for stringNo in range(len(self._vibrating)):
            self.endVibration(stringNo, offset)



    def fretboardNote(self, stringNo, chordList, barreChordFretNo):
        openString = self._tuning[stringNo] + self.capoFretNo + barreChordFretNo
        lowestFretNo = 127
        for note in chordList:
            fretNo = note - openString % 12
            while fretNo < 0:
                fretNo += 12
            # find the lowest note
            if fretNo < lowestFretNo:
                lowestFretNo = fretNo

        return openString + lowestFretNo



    def trackBar(self, pattern, ctable):
        """ Do the bass bar.

        Called from self.bar()

        """

        sc = self.seq
        unify = self.unify[sc]


        for p in pattern:
            try:
                ct = self.getChordInPos(p.offset, ctable)
                chordList=ct.chord.noteList     # catch the case when there is no noteList attribute
            except AttributeError:
                continue

            # Todo
            #if ct.bassZ:
            #    continue

            if len(self._tuning) != len(p.pluckVol):
                error("Wrong number of strings")
                continue

            if gbl.debug:
                self.printChordShape(ct)

            plectrumNoteOnList = [] # for debugging only
            previousNote = -1


            # Find how many strings have been plucked this time
            pluckStringCount = 0;
            for vol in p.pluckVol:
                if vol == -1:
                    continue
                pluckStringCount += 1


            pluckStringIndex = 0;
            for stringNo, vol in enumerate(p.pluckVol):
                if vol == -1:
                    continue

                # find the centre of the strum
                strumOffset = p.offset + p.strum*(pluckStringIndex - pluckStringCount/2.0);
                pluckStringIndex += 1

                self.endVibration(stringNo, strumOffset)
                if vol >= 1:
                    note = self.fretboardNote(stringNo, chordList, 0)

                    if note == previousNote:
                        warn("duplicate note in the chord Igrnoring")
                        continue
                    previousNote = note

                    outputVolume = self.adjustVolume(vol, p.offset)

                    gbl.mtrks[self.channel].addNoteOnToTrack(strumOffset, note, outputVolume)

                    self._vibrating[stringNo].note = note
                    if outputVolume == 0:
                        self._vibrating[stringNo].note = -1

                    plectrumNoteOnList.append(note)  # for debugging only


            if gbl.debug:
                print  "    channel ", self.channel, "offset", p.offset, "chordList ", chordList,  " NoteOn ", plectrumNoteOnList



    def printChordShape(self, chordTable, barreChordFretNo = 0):
        chordList=chordTable.chord.noteList     # catch the case when there is no noteList attribute

        if hasattr(self,  'previousChordList'):
            if chordList == self.previousChordList:
                return
        self.previousChordList = chordList

        previousNote = -1
        print
        print "       Plectrum ", chordTable.name, " chord"
        for stringNo, openNote in enumerate(reversed(self._tuning)):

            note = self.fretboardNote(len(self._tuning) - stringNo -1, chordList, barreChordFretNo)

            if note == previousNote:
                warn("duplicate note in the chord Igrnoring")
                continue
            previousNote = note

            print "   ", openNote, "|",
            i=-1
            for x in range (note - openNote -1):
                print "-",
                i = x
            if note != openNote:
                print "*",
                i +=1
            for j in range (12 - i):
                print "-",
            print "-", note
        print



def noteNameToMidiPitch(s):
    """ Convert a name ('e', 'g#') to a MIDI pitch. """

    tb = { 'c': 0,  'c#': 1,  'd&': 1, 'd': 2, 'd#': 3,  'e&': 3,
           'e': 4,  'f&': 4,  'e#': 5, 'f': 5, 'f#': 6,  'g&': 6,
           'g': 7,  'g#': 8,  'a&': 8, 'a': 9, 'a#': 10, 'b&': 10,
           'b': 11, 'b&': 11, 'c&': 11,  'b#': 0 }

    # strip and count trailing '+' and '-'

    if '-' in s and '+' in s:
        error("Can't have both + and - in note name.")

    adjust = 0
    while s.endswith('-'):
        adjust -= 12
        s=s[:-1]
    while s.endswith('+'):
        adjust += 12
        s=s[:-1]

    try:
        value = tb[s] + 60  # puts into middle octave 60==5*12
    except:
        error ( "No such name: %s" % s)

    return value + adjust





